#!/usr/bin/env python
#conversion_test.py
#Randomization test for gene conversion rate and confidence interval calculation
#Program reads a single FASTA file. For multiple gene analyses (e.g. 16S and 23S) concatenated alignments of each operon are used
#This is version 3, which takes into account multiple successive conversions within a single operon and uses recursion to identify the fewest conversion events required to produce the final sequences
#@author: Isaac Gifford

#Sample command line input
#python3 conversion_test.py -r 10000 -s conversion_distance.txt -f ecoli_rRNA_aligned.fasta -i 15 -l .00030 -u .00038 -t .000005 -m rRNA -o sample

import argparse
import sys
import os
import random
import numpy
import itertools
from scipy.stats import poisson
from statistics import mean

parser = argparse.ArgumentParser(description="Read in conversion size and FASTA files, generate random conversions, and calculate statistics.")
parser.add_argument("-r", "--replicates", type=int, default=10000, help="number of replicates")
parser.add_argument("-s", "--sizefile", help='tab-delimited file containing min and max sizes of observed conversions')
parser.add_argument("-f", "--fastafile", help='name of FASTA file for operon sequences')
parser.add_argument("-i", "--lineages", type=int, help="number of lineages")
parser.add_argument("-l", "--lowestrate", help="lowest rate to test")
parser.add_argument("-u", "--upperrate", help="highest rate to test")
parser.add_argument("-t", "--step", help="step size of rates to test")
parser.add_argument("-m", "--mode", help="rRNA, 16S, 23S, or intergenic mode")
parser.add_argument("-o", "--output", help="name of output file")

args = parser.parse_args() 

######################################## Functions ###################################################

#Separates converted sites into region "chunks" based on shared ancestral sequences
def chunkifier (site_list, donor_list, removed_list, chunks_list):
	conversion_counts = {}
	for i in range(0, len(ancestor_seqs)):
		conversion_counts[i] = 0
	for i in site_list:
		for j in donor_list[i]:
			conversion_counts[j] = conversion_counts[j] + 1
	successful = []
	for i in conversion_counts:
		if (conversion_counts[i] == len(site_list)):
			successful.append(i)
	if (len(successful) > 0):
		site_list = removed_list
		removed_list = []
		chunks_list.append(successful)
		if (len(site_list) > 0):
			return chunkifier(site_list, donor_list, removed_list, chunks_list)
	else:
		removed_list.insert(0, site_list[len(site_list)-1])
		del site_list[len(site_list)-1]
		return chunkifier(site_list, donor_list, removed_list, chunks_list)

#identifies shared ancestral sequences between chunks
def simplifier (chunks_list):
	arrangement_chunks = {}
	arrangements = list(itertools.product(*chunks_list))
	for i in arrangements:
		different_chunks = {}
		for j in i:
			different_chunks[j] = 0
		number_of_different_chunks = len(different_chunks)
		arrangement_chunks[i] = number_of_different_chunks
	
	best_arrangement = random.choice(list(arrangement_chunks.keys()))
	different_chunks = arrangement_chunks[best_arrangement]
	
	for i in arrangement_chunks:
		if (arrangement_chunks[i] < different_chunks):
			different_chunks = arrangement_chunks[i]
	best_arrangements = []
	for i in arrangement_chunks:
		if (arrangement_chunks[i] == different_chunks):
			best_arrangements.append(i)
	return(best_arrangements)

#finds the simplest group of ancestral sequences for all chunks
def parsimony (simplified_chunks, conversions, holder, chunks_holder):
	if (len(chunks_holder) > 0):
		passed_chunks = chunks_holder
		chunks_holder = []
		for i in passed_chunks:
			current_chunks = list(i)
			separate_chunks = {}
			for j in current_chunks:
				separate_chunks[j] = 0
			for j in current_chunks:
				separate_chunks[j] = separate_chunks[j] + 1
			if (len(separate_chunks) == len(i)):
				conversions = conversions + len(separate_chunks)
			else:
				chunks_max = 0
				max_chunk = ''
				for j in separate_chunks:
					chunks_max = separate_chunks[j]
					max_chunk = j
		
				conversions = conversions + 1
		
				building = []
				
				for j in i:
					if (j == max_chunk):
						chunks_holder.append(building)
						building = []
					else:
						building.append(j)
		
				return parsimony(current_chunks, conversions, holder, chunks_holder)
		return int(conversions)
	elif (len(holder) == 0):
		if (len(simplified_chunks) == 0):
			return int(0)
		current_chunks = simplified_chunks
		separate_chunks = {}
		for i in simplified_chunks:
			separate_chunks[i] = 0
		for i in simplified_chunks:
			separate_chunks[i] = separate_chunks[i] + 1
		if (len(separate_chunks) == len(simplified_chunks)):
			conversions = conversions + len(separate_chunks)
			if (len(holder)>0):
				simplified_chunks = holder
				holder = []
				return parsimony(current_chunks, conversions, holder, chunks_holder)
			else:
				return int(conversions)
				holder = []
				conversions = 0
		else:
			chunks_max = 0
			max_chunk = ''
			for j in separate_chunks:
				if (separate_chunks[j] > chunks_max):
					chunks_max = separate_chunks[j]
					max_chunk = j
			conversions = conversions + 1

			building = []
			for j in simplified_chunks:
				if (j == max_chunk):
					if (len(building)>0):
						chunks_holder.append(building)
					building = []
				else:
					building.append(j)
			chunks_holder.append(building)											#appends leftover chunks
	
			return parsimony(current_chunks, conversions, holder, chunks_holder)

#calculates the number of conversions required to produce the input sequence
def conversion_counter(ancestors, sites_converted, ancestor_number, evolved_seq, sites_counter):
	sites_converted_temp = []
	for i in sites_converted:
		sites_converted_temp.append(i)
	sites_converted_temp.sort()

	potential_donors = {}

	for j in (sites_converted_temp):
		potential_donors_array = []
		for k in range(0, len(ancestors)):
		#for k in ancestor_seqs:
			compare_header = ancestor_headers[k]
			compare_seq = ancestors[k]
			if (compare_seq[j-1] == evolved_seq[j-1]):
				potential_donors_array.append(k)
	
		potential_donors[j] = potential_donors_array
	
	#removing ancestral positions from ends
	sites_to_delete = []
	for k in range(0, len(sites_converted_temp)):												#beginning from the beginning forward
		matched = 0
		for j in potential_donors[sites_converted_temp[k]]:
			if (j == ancestor_number):
				matched = 1
		if (matched == 1):
			sites_to_delete.append(k)
		if (matched == 0):
			break

	for k in range(len(sites_converted_temp)-1, 0, -1):												#searches from the end in reverse
		matched = 0
		for j in potential_donors[sites_converted_temp[k]]:
			if (j == ancestor_number):
				matched = 1
		if (matched == 1):
			sites_to_delete.append(k)
		if (matched == 0):
			break

	duplicate_check = {}
	for k in sites_to_delete:
		duplicate_check[k] = 0
	duplicate_free = []
	for k in duplicate_check:
		duplicate_free.append(k)
	duplicate_free.sort()

	for k in range(len(duplicate_free), 0, -1):
		del sites_converted_temp[duplicate_free[k-1]]

	#identify parental chunks
	potential_conversions = len(potential_donors)
	removed = []
	sites = []
	chunks = []
	if (len(sites_converted) > 0):
		for i in sites_converted_temp:
			sites.append(i)
			binary_ancestor = 0
			for j in potential_donors[i]:
				if (j != ancestor_number or len(potential_donors[i]) > 1):
					binary_ancestor = 1
			if (binary_ancestor == 1):
				sites_counter[i] = sites_counter[i] + 1											#records heterologous site conversions
		chunkifier(sites, potential_donors, removed, chunks)

	if (len(chunks) > 0):
		#count observable conversions
		different_chunks = {}
		different_chunks[1] = chunks[0]
	
		simplified = simplifier(chunks)
		#print(simplified)
	
		simplified_conversions = []
	
		for j in simplified:
			total_conversions = 0
			checked_chunks = []
			building_selfcheck = []
			for k in j:
				if (k == ancestor_number):
					checked_chunks.append(building_selfcheck)
					building_selfcheck = []
				else:
					building_selfcheck.append(k)
			checked_chunks.append(building_selfcheck)
			
			for k in checked_chunks:
				conversions_counting = 0
				chunk_holder = []
				chunk_of_chunks = []
				parsimonious = parsimony(k, conversions_counting, chunk_holder, chunk_of_chunks)
				total_conversions = total_conversions + parsimonious
			simplified_conversions.append(total_conversions)
	
		optimized = min(simplified_conversions)
		return optimized
	elif (len(chunks) == 0):
		total_conversions = 0
		return total_conversions
###########################################################################################

if not os.path.isdir(os.path.dirname(args.output)):
	os.makedirs(os.path.dirname(args.output))

output_name = args.output + '.rates.txt'
output = open(output_name, 'w')

headers_seqs = {}
operons = []
seq_lengths = []
ancestor_headers = []
ancestor_seqs = []

operons_converted = {}

#read alignment sequences
with open(args.fastafile) as fastafile:
	header = ""
	line_count = 2
	for i in fastafile:
		i = i.rstrip('\n')
		if (line_count%2==0):
			header = i
			ancestor_headers.append(header)
			operons.append(i)
			operons_converted[i] = []
		else:
			headers_seqs[header] = i
			seq_lengths.append(len(i))
			ancestor_seqs.append(i)
		line_count = line_count + 1

#identify heterologous sites
sites_converted = {}
sites_converted_list = []
seq_length = seq_lengths[0]
for i in seq_lengths:
	if (int(i) != int(seq_length)):
		sys.exit("Alignment lengths are not equal")
print("Identifying heterologous sites")
for i in range(0, seq_length):
	temp_bases = []
	binary = 0
	for j in headers_seqs:
		seq = headers_seqs[j]
		temp_bases.append(seq[i])
	for k in temp_bases:
		if (k != temp_bases[0]):
			binary = 1
	if (binary == 1):
		alignment_position = int(i) + 1
		sites_converted[alignment_position] = []
		sites_converted_list.append(alignment_position)

replicates = args.replicates
observables = []
unobservables = []

rates = numpy.arange(float(args.lowestrate), float(args.upperrate), float(args.step))

#read conversion file
sizes = []
with open(args.sizefile) as sizefile:
	for i in sizefile:
		i=i.rstrip('\n')
		sizes.append(i)

rates_results = {}

lineages = args.lineages

max_rate = 0
max_successes = 0
conversions_per_operon = {}
conversions_per_site = {}

print()
print("Testing candidate rates")
for r in rates:
	rate = round(r, 7)
	successful = 0
	
	conversions_per_operon_rate = {}
	for i in ancestor_headers:
		conversions_per_operon_rate[i] = []
	conversions_per_site_rate = {}
	for i in sites_converted_list:
		conversions_per_site_rate[i] = []
	
	for replicate in range(0, replicates):
		if (replicate%500==0):
			print("Running "+str(rate)+" replicate "+str(replicate+1))
		
		lineage_seqs = {}
		for i in range(1, lineages + 1):
			for j in headers_seqs:
				header = str(i) + j
				lineage_seqs[header] = headers_seqs[j]
		
		unobservable = 0
		observable = 0
		observable_16 = 0
		observable_23 = 0
		observable_inter = 0
		conversions_performed = 0
		
		conversions_per_operon_temp = {}
		for i in ancestor_headers:
			conversions_per_operon_temp[i] = 0
		conversions_per_site_temp = {}
		for i in sites_converted_list:
			conversions_per_site_temp[i] = 0
	
		poisson_mean = r * 13750 * lineages														#for 13750 generations
		total_conversions = poisson.rvs(poisson_mean)
		for converts in range(0, total_conversions):
			#pick random conversion
			conversion = random.randrange(0, len(sizes))
			splitted = sizes[conversion].split("\t")
			gene = splitted[0]
		
			#generate random size
			if (args.mode == 'rRNA' or args.mode == '16S' or args.mode == '23S'):
				if (int(splitted[1])>50):
					rand_size = random.randrange(int(splitted[1]), int(splitted[2]))
				else:
					rand_size = random.randrange(50, int(splitted[2]))
			elif (args.mode == 'intergenic'):
				rand_size = 188
		
			#generate random start location and end location
			if (args.mode == 'rRNA' or args.mode == '16S' or args.mode == '23S'):
				if (gene == '16'):
					rand_start = random.randrange(1, 1543-rand_size)
				elif (gene == '23'):
					rand_start = random.randrange(1544, 4576-rand_size)
			elif (args.mode == 'intergenic'):
				rand_start = 1
			end = rand_start + rand_size
		
			#generate donor and recipient operons
			conversion_lineage = random.randrange(1, lineages)
			donor = random.choices(operons)
			recipient = random.choices(operons)
			while (donor == recipient):
				recipient = random.choices(operons)
		
			#get donor and recipient seqs
			donor_seq = lineage_seqs[str(conversion_lineage)+donor[0]]
			donor_fragment = donor_seq[rand_start:end]
			recipient_seq=lineage_seqs[str(conversion_lineage)+recipient[0]]
			recipient_fragment = recipient_seq[rand_start:end]
			
			#compare donor and recipient fragments
			if (donor_fragment == recipient_fragment):
				unobservable = unobservable + 1
			else:
				if (args.mode == 'rRNA' or args.mode == '16S' or args.mode == '23S'):
					if (gene == '16'):
						before_fragment = recipient_seq[0:rand_start]
						after_fragment = recipient_seq[end:4576]
						lineage_key = str(conversion_lineage) + recipient[0]
						lineage_seqs[lineage_key] = before_fragment + donor_fragment + after_fragment
					elif (gene == '23'):
						before_fragment = recipient_seq[0:rand_start]
						after_fragment = recipient_seq[end:4576]
						lineage_key = str(conversion_lineage) + recipient[0]
						lineage_seqs[lineage_key] = before_fragment + donor_fragment + after_fragment
				elif (args.mode == 'intergenic'):
					before_fragment = recipient_seq[0:rand_start]
					after_fragment = recipient_seq[end:189]
					lineage_seqs[str(conversion_lineage)+str(recipient[0])] = before_fragment + donor_fragment + after_fragment
		
		#count conversions in final sequences
		conversions_16S = 0
		conversions_23S = 0
		inter_conversions = 0
		
		for l in lineage_seqs:
			ancestor_split = l.split(">")
			ancestor_name = ">" + ancestor_split[1]
			ancestor_number = 999
			for m in range(0, len(ancestor_headers)):
				if (ancestor_headers[m] == ancestor_name):
					ancestor_number = m
			
			if (args.mode == 'rRNA' or args.mode == '16S'):
				ancestor_16S = []
				sites_16S = []
				for j in sites_converted_list:
					if (j < 1543):
						sites_16S.append(j)
				
				counted_16S = conversion_counter(ancestor_seqs, sites_16S, int(ancestor_number), lineage_seqs[l], conversions_per_site_temp)
				conversions_16S = conversions_16S + counted_16S
				conversions_per_operon_temp[ancestor_name] = conversions_per_operon_temp[ancestor_name] + counted_16S
			
			if (args.mode == 'rRNA' or args.mode == '23S'):
				ancestor_23S = []
				sites_23S = []
				for j in sites_converted_list:
					if (j > 1542):
						sites_23S.append(j)
				
				counted_23S = conversion_counter(ancestor_seqs, sites_23S, int(ancestor_number), lineage_seqs[l], conversions_per_site_temp)
				conversions_23S = conversions_23S + counted_23S
				conversions_per_operon_temp[ancestor_name] = conversions_per_operon_temp[ancestor_name] + counted_23S
			
			if (args.mode == 'intergenic'):
				ancestor_inter = []
				sites_inter = []
				for j in sites_converted_list:
					sites_inter.append(j)
				
				counted_inter = conversion_counter(ancestor_seqs, sites_inter, int(ancestor_number), lineage_seqs[l], conversions_per_site_temp)
				inter_conversions = inter_conversions + counted_inter

		for j in conversions_per_operon_temp:
			conversions_per_operon_rate[j].append(conversions_per_operon_temp[j])
		for j in conversions_per_site_temp:
			conversions_per_site_rate[j].append(conversions_per_site_temp[j])
		
		if (args.mode == 'rRNA'):
			if (conversions_16S == 7 and conversions_23S == 27):
				successful = successful + 1
		elif (args.mode == '16S'):
			if (conversions_16S == 7):
				successful = successful + 1
		elif (args.mode == '23S'):
			if (conversions_23S == 27):
				successful = successful + 1
		elif (args.mode == 'intergenic'):
			if (inter_conversions == 4):
				successful = successful + 1
		
		if (successful > max_successes):
			max_rate = rate
			max_successes = successful
			for j in conversions_per_operon_rate:
				conversions_per_operon[j] = conversions_per_operon_rate[j]
			for j in conversions_per_site_rate:
				conversions_per_site[j] = conversions_per_site_rate[j]
	
	rates_results[rate] = successful
	print("Rate: "+str(rate)+"\tSuccesses: "+str(successful))

print("Results", file = output)
print("Tested rate\tNumber of successes", file = output)
for i in rates_results:
	print(str(i)+"\t"+str(rates_results[i]), file = output)

print()
print("Optimal rate: "+str(max_rate))
print()
print("\n\nOptimal conversion rate: "+str(max_rate), file = output)

#Determine average conversions per operon and confidence intervals
if (args.mode == 'rRNA'):
	print()
	print("Calculating conversions per operon")
	print("Operon\tAverage conversions\tLower bound\tUpper bound")
	for i in conversions_per_operon:
		per_operon = conversions_per_operon[i]
		per_operon.sort()
		lower_position = int(replicates * 0.025)
		lower_bound = per_operon[lower_position]
		upper_position = int(replicates * 0.975)
		upper_bound = per_operon[upper_position]
	
		sum = 0
		for j in conversions_per_operon[i]:
			sum = sum + int(j)
		average = sum / replicates
	
		print(str(i)+"\t"+str(average)+"\t"+str(lower_bound)+"\t"+str(upper_bound))
		print(str(i)+"\t"+str(average)+"\t"+str(lower_bound)+"\t"+str(upper_bound), file = output)

#Determine average conversions per heterologous site and confidence intervals
if (args.mode == 'rRNA'):
	print()
	print("Calculating conversions per operon")
	print("Site\tAverage conversions\tLower bound\tUpper bound")
	for i in conversions_per_site:
		per_site = conversions_per_site[i]
		per_site.sort()
		lower_position = int(replicates * 0.025)
		lower_bound = per_site[lower_position]
		upper_position = int(replicates * 0.975)
		upper_bound = per_site[upper_position]
	
		sum = 0
		for j in conversions_per_site[i]:
			sum = sum + int(j)
		average = sum / replicates
	
		print(str(i)+"\t"+str(average)+"\t"+str(lower_bound)+"\t"+str(upper_bound))
		print(str(i)+"\t"+str(average)+"\t"+str(lower_bound)+"\t"+str(upper_bound), file = output)

#Calculates lower confidence interval
binary = 0
lower_rate = max_rate
print()
print("Calculating lower bound")
while (binary!=1):
	successful = 0
	for replicate in range(0, replicates):
		lineage_seqs = {}
		for i in range(1, lineages + 1):
			for j in headers_seqs:
				header = str(i) + j
				lineage_seqs[header] = headers_seqs[j]
		
		unobservable = 0
		observable = 0
		observable_16 = 0
		observable_23 = 0
		observable_inter = 0
		conversions_performed = 0
		
		conversions_per_operon_temp = {}
		for i in ancestor_headers:
			conversions_per_operon_temp[i] = 0
		conversions_per_site_temp = {}
		for i in sites_converted_list:
			conversions_per_site_temp[i] = 0
	
		conversions = poisson.rvs(lower_rate * lineages * 13750)
		for converts in range(0, conversions):
			#pick random conversion
			conversion = random.randrange(0, len(sizes))
			splitted = sizes[conversion].split("\t")
			gene = splitted[0]
		
			#generate random size
			if (args.mode == 'rRNA' or args.mode == '16S' or args.mode == '23S'):
				if (int(splitted[1])>50):
					rand_size = random.randrange(int(splitted[1]), int(splitted[2]))
				else:
					rand_size = random.randrange(50, int(splitted[2]))
			elif (args.mode == 'intergenic'):
				rand_size = 188
		
			#generate random start location and end location
			if (args.mode == 'rRNA' or args.mode == '16S' or args.mode == '23S'):
				if (gene == '16'):
					rand_start = random.randrange(1, 1543-rand_size)
				elif (gene == '23'):
					rand_start = random.randrange(1544, 4576-rand_size)
			elif (args.mode == 'intergenic'):
				rand_start = 1
			end = rand_start + rand_size
		
			#generate donor and recipient operons
			conversion_lineage = random.randrange(1, lineages)
			donor = random.choices(operons)
			recipient = random.choices(operons)
			while (donor == recipient):
				recipient = random.choices(operons)
		
			#get donor and recipient seqs
			donor_seq = lineage_seqs[str(conversion_lineage)+donor[0]]
			donor_fragment = donor_seq[rand_start:end]
			recipient_seq=lineage_seqs[str(conversion_lineage)+recipient[0]]
			recipient_fragment = recipient_seq[rand_start:end]
			
			#compare donor and recipient fragments
			if (donor_fragment == recipient_fragment):
				unobservable = unobservable + 1
			else:
				if (args.mode == 'rRNA' or args.mode == '16S' or args.mode == '23S'):
					if (gene == '16'):
						before_fragment = recipient_seq[0:rand_start]
						after_fragment = recipient_seq[end:4576]
						lineage_key = str(conversion_lineage) + recipient[0]
						lineage_seqs[lineage_key] = before_fragment + donor_fragment + after_fragment
					elif (gene == '23'):
						before_fragment = recipient_seq[0:rand_start]
						after_fragment = recipient_seq[end:4576]
						lineage_key = str(conversion_lineage) + recipient[0]
						lineage_seqs[lineage_key] = before_fragment + donor_fragment + after_fragment
				elif (args.mode == 'intergenic'):
					before_fragment = recipient_seq[0:rand_start]
					after_fragment = recipient_seq[end:189]
					lineage_seqs[str(conversion_lineage)+str(recipient[0])] = before_fragment + donor_fragment + after_fragment
		
		#count conversions in final sequences
		conversions_16S = 0
		conversions_23S = 0
		inter_conversions = 0
		
		for l in lineage_seqs:
			ancestor_split = l.split(">")
			ancestor_name = ">" + ancestor_split[1]
			ancestor_number = 999
			for m in range(0, len(ancestor_headers)):
				if (ancestor_headers[m] == ancestor_name):
					ancestor_number = m
			
			if (args.mode == 'rRNA' or args.mode == '16S'):
				ancestor_16S = []
				sites_16S = []
				for j in sites_converted_list:
					if (j < 1543):
						sites_16S.append(j)
				
				counted_16S = conversion_counter(ancestor_seqs, sites_16S, int(ancestor_number), lineage_seqs[l], conversions_per_site_temp)
				conversions_16S = conversions_16S + counted_16S
				conversions_per_operon_temp[ancestor_name] = conversions_per_operon_temp[ancestor_name] + counted_16S
			
			if (args.mode == 'rRNA' or args.mode == '23S'):
				ancestor_23S = []
				sites_23S = []
				for j in sites_converted_list:
					if (j > 1542):
						sites_23S.append(j)
				
				counted_23S = conversion_counter(ancestor_seqs, sites_23S, int(ancestor_number), lineage_seqs[l], conversions_per_site_temp)
				conversions_23S = conversions_23S + counted_23S
				conversions_per_operon_temp[ancestor_name] = conversions_per_operon_temp[ancestor_name] + counted_23S
			
			if (args.mode == 'intergenic'):
				ancestor_inter = []
				sites_inter = []
				for j in sites_converted_list:
					sites_inter.append(j)
				
				counted_inter = conversion_counter(ancestor_seqs, sites_inter, int(ancestor_number), lineage_seqs[l], conversions_per_site_temp)
				inter_conversions = inter_conversions + counted_inter
		
		if (args.mode == 'rRNA'):
			if (conversions_16S >= 7 and conversions_23S >= 27):
				successful = successful + 1
		elif (args.mode == '16S'):
			if (conversions_16S >= 7):
				successful = successful + 1
		elif (args.mode == '23S'):
			if (conversions_23S >= 27):
				successful = successful + 1
		elif (args.mode == 'intergenic'):
			if (inter_conversions >= 4):
				successful = successful + 1
	lower_rounded = round(lower_rate, 8)
	print("Rate: "+str(lower_rounded)+"\tSuccesses: "+str(successful))
	if (successful == (replicates * 0.025)):
		binary = 1
	elif (successful > (replicates * 0.025)):
		lower_rate = lower_rate - float(args.step)
	elif (successful < (replicates * 0.025)):
		lower_rate = conversions + float(args.step)
		binary = 1
lower_rate = lower_rounded

#calculates upper confidence interval
binary = 0
upper_rate = max_rate
print()
print("Calculating upper bound")
while (binary!=1):
	successful = 0
	for replicate in range(0, replicates):
		unobservable = 0
		
		conversions = poisson.rvs(upper_rate * lineages * 13750)
		
		lineage_seqs = {}
		for i in range(1, lineages + 1):
			for j in headers_seqs:
				header = str(i) + j
				lineage_seqs[header] = headers_seqs[j]
		
		for converts in range(0, conversions):
			#pick random conversion
			conversion = random.randrange(0, len(sizes))
		
			#generate random size
			if (args.mode == 'rRNA' or args.mode == '16S' or args.mode == '23S'):
				if (int(splitted[1])>50):
					rand_size = random.randrange(int(splitted[1]), int(splitted[2]))
				else:
					rand_size = random.randrange(50, int(splitted[2]))
			elif (args.mode == 'intergenic'):
				rand_size = 188
		
			gene = splitted[0]
		
			#generate random start location and end location
			if (args.mode == 'rRNA' or args.mode == '16S' or args.mode == '23S'):
				if (gene == '16'):
					rand_start = random.randrange(1, 1543-rand_size)
				elif (gene == '23'):
					rand_start = random.randrange(1544, 4576-rand_size)
			elif (args.mode == 'intergenic'):
				rand_start = 1
			end = rand_start + rand_size
		
			#generate donor and recipient operons
			conversion_lineage = random.randrange(1, lineages)
			donor = random.choices(operons)
			recipient = random.choices(operons)
			while (donor == recipient):
				recipient = random.choices(operons)
		
			#get donor and recipient seqs
			donor_seq = lineage_seqs[str(conversion_lineage)+donor[0]]
			donor_fragment = donor_seq[rand_start:end]
			recipient_seq=lineage_seqs[str(conversion_lineage)+recipient[0]]
			recipient_fragment = recipient_seq[rand_start:end]
			
			#compare donor and recipient fragments
			if (donor_fragment == recipient_fragment):
				unobservable = unobservable + 1
			else:
				if (args.mode == 'rRNA' or args.mode == '16S' or args.mode == '23S'):
					if (gene == '16'):
						before_fragment = recipient_seq[0:rand_start]
						after_fragment = recipient_seq[end:4576]
						lineage_key = str(conversion_lineage) + recipient[0]
						lineage_seqs[lineage_key] = before_fragment + donor_fragment + after_fragment
					elif (gene == '23'):
						before_fragment = recipient_seq[0:rand_start]
						after_fragment = recipient_seq[end:4576]
						lineage_key = str(conversion_lineage) + recipient[0]
						lineage_seqs[lineage_key] = before_fragment + donor_fragment + after_fragment
				elif (args.mode == 'intergenic'):
					before_fragment = recipient_seq[0:rand_start]
					after_fragment = recipient_seq[end:189]
					lineage_seqs[str(conversion_lineage)+str(recipient[0])] = before_fragment + donor_fragment + after_fragment
		
		#count conversions in final sequences
		conversions_16S = 0
		conversions_23S = 0
		inter_conversions = 0
		
		for l in lineage_seqs:
			ancestor_split = l.split(">")
			ancestor_name = ">" + ancestor_split[1]
			ancestor_number = 999
			for m in range(0, len(ancestor_headers)):
				if (ancestor_headers[m] == ancestor_name):
					ancestor_number = m
			
			if (args.mode == 'rRNA' or args.mode == '16S'):
				ancestor_16S = []
				sites_16S = []
				for j in sites_converted_list:
					if (j < 1543):
						sites_16S.append(j)
				
				counted_16S = conversion_counter(ancestor_seqs, sites_16S, int(ancestor_number), lineage_seqs[l], conversions_per_site_temp)
				conversions_16S = conversions_16S + counted_16S
				conversions_per_operon_temp[ancestor_name] = conversions_per_operon_temp[ancestor_name] + counted_16S
			if (args.mode == 'rRNA' or args.mode == '23S'):
				ancestor_23S = []
				sites_23S = []
				for j in sites_converted_list:
					if (j > 1542):
						sites_23S.append(j)
				
				counted_23S = conversion_counter(ancestor_seqs, sites_23S, int(ancestor_number), lineage_seqs[l], conversions_per_site_temp)
				conversions_23S = conversions_23S + counted_23S
				conversions_per_operon_temp[ancestor_name] = conversions_per_operon_temp[ancestor_name] + counted_23S
			if (args.mode == 'intergenic'):
				ancestor_inter = []
				sites_inter = []
				for j in sites_converted_list:
					sites_inter.append(j)
				
				counted_inter = conversion_counter(ancestor_seqs, sites_inter, int(ancestor_number), lineage_seqs[l], conversions_per_site_temp)
				inter_conversions = inter_conversions + counted_inter

		if (args.mode == 'rRNA'):
			if (conversions_16S <= 7 and conversions_23S <= 27):
				successful = successful + 1
		elif (args.mode == '16S'):
			if (conversions_16S <= 7):
				successful = successful + 1
		elif (args.mode == '23S'):
			if (conversions_23S <= 27):
				successful = successful + 1
		elif (args.mode == 'intergenic'):
			if (inter_conversions <= 4):
				successful = successful + 1

	upper_rounded = round(upper_rate, 8)
	print("Rate: "+str(upper_rounded)+"\tSuccesses: "+str(successful))
	if (successful == (replicates * 0.025)):
		binary = 1
	elif (successful > (replicates * 0.025)):
		upper_rate = upper_rate + float(args.step)
	elif (successful < (replicates * 0.025)):
		upper_rate = conversions - float(args.step)
		binary = 1
upper_rate = upper_rounded

print()
print("Lower bound rate: "+str(lower_rate))
print("Upper bound rate: "+str(upper_rate))
print("Lower bound rate: "+str(lower_rate), file = output)
print("Upper bound rate: "+str(upper_rate), file = output)

output.close()